/**
 * @file artis/kernel/AbstractCoupledModel.hpp
 * @author See the AUTHORS file
 */

/*
 * Copyright (C) 2012-2016 ULCO http://www.univ-littoral.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __ARTIS_KERNEL_ABSTRACT_COUPLED_MODEL_HPP
#define __ARTIS_KERNEL_ABSTRACT_COUPLED_MODEL_HPP

#include <artis/kernel/AbstractModel.hpp>
#include <artis/kernel/Externals.hpp>
#include <artis/kernel/Internals.hpp>
#include <artis/kernel/States.hpp>

#include <artis/utils/DateTime.hpp>
#include <artis/utils/Trace.hpp>

#include <functional>
#include <vector>

namespace artis { namespace kernel {

template < typename T, typename U, typename V, typename W >
class AbstractCoupledModel : public AbstractModel < U, V >,
                             public States < T, U >,
                             public Internals < T, U >,
                             public Externals < T, U >
{
    typedef AbstractModel < U, V > type;
    typedef std::map < int, type* > Submodels;
    typedef std::map < int,
                       std::vector < type* > > Setsubmodels;
    typedef std::map < int, std::pair < type*,
                                        int > > SubModelInternals;

    struct Var
    {
        unsigned int index;
        type* model;
        int sub_index;
    };

public:
    AbstractCoupledModel(const type* parent = 0) : type(parent)
    { }

    virtual ~AbstractCoupledModel()
    { }

    virtual void after(typename U::type t)
    {

#ifdef WITH_TRACE
        utils::Trace < utils::DoubleTime >::trace()
            << utils::TraceElement < utils::DoubleTime >("KERNEL", t,
                                                         utils::KERNEL)
            << " - AFTER - "
            << AbstractCoupledModel < T, U, V, W >::path(this);
        utils::Trace < utils::DoubleTime >::trace().flush();
#endif

    }

    virtual const Node < U >* atomic(unsigned int index) const
    {
        typename AbstractCoupledModel::SubModelInternals::const_iterator it =
            submodel_internals.find(index);

        if (it == submodel_internals.end()) {
            return this;
        } else {
            return it->second.first->atomic(it->second.second);
        }
    }

    virtual void before(typename U::type t)
    {

#ifdef WITH_TRACE
        utils::Trace < utils::DoubleTime >::trace()
            << utils::TraceElement < utils::DoubleTime >("KERNEL", t,
                                                         utils::KERNEL)
            << " - BEFORE - "
            << AbstractCoupledModel < T, U, V, W >::path(this);
        utils::Trace < utils::DoubleTime >::trace().flush();
#endif

    }

    virtual void build(const W& /* parameters */)
    { }

    virtual void build(const W& /* parameters */,
                       const std::string& /* json */)
    { }

    virtual bool check(typename U::type t) const
    { return Externals < T, U >::check(t); }

    virtual void compute(typename U::type t, bool update) = 0;

    virtual const Any& get(typename U::type t, unsigned int index) const
    {
        typename AbstractCoupledModel::SubModelInternals::const_iterator it =
            submodel_internals.find(index);

        if (it == submodel_internals.end()) {
            if (type::last_time != t) {
                throw utils::InvalidGet("Variable not computed");
            }
            return Internals < T, U >::get(index);
        } else {
            return it->second.first->get(t, it->second.second);
        }
    }

    template < typename Z, typename K >
    Z get(typename U::type t, unsigned int index) const
    {
        typename AbstractCoupledModel::SubModelInternals::const_iterator it =
            submodel_internals.find(index);

        if (it == submodel_internals.end()) {
            if (type::last_time != t) {
                throw utils::InvalidGet("Variable not computed");
            }
            const Any& value = Internals < T, U >::get(index);

            return static_cast < const T* >(this)->*(value.get < T, Z >());
        } else {
            const Any& value = it->second.first->get(t, it->second.second);
            const Node < U >* object = it->second.first->atomic(
                it->second.second);

            return static_cast < const K* >(object)->*(value.get < K, Z >());
        }
    }

    virtual std::string get(const ValueType& value_type, typename U::type t,
                            unsigned int index) const
    {
        typename AbstractCoupledModel::SubModelInternals::const_iterator it =
            submodel_internals.find(index);

        if (it == submodel_internals.end()) {
            Any value = Internals < T, U >::get(index);

            switch (value_type) {
            case DOUBLE:
                return std::to_string(
                    static_cast < const T* >(this)->*(value.get < T,
                                                      double >()));
            case INT:
                return std::to_string(
                    static_cast < const T* >(this)->*(value.get < T, int >()));
            case BOOL:
                return std::to_string(
                    static_cast < const T* >(this)->*(value.get < T, bool >()));
            default: return "NA";
            }
        } else {
            return it->second.first->get(value_type, t, it->second.second);
        }
    }

    virtual void init(typename U::type t, const V& parameters) = 0;

    virtual bool is_atomic() const
    { return false; }

    bool is_computed(typename U::type t, unsigned int index) const
    {
        typename AbstractCoupledModel::SubModelInternals::const_iterator it =
            submodel_internals.find(index);

        if (it == submodel_internals.end()) {
            return false;
        } else {
            return it->second.first->is_computed(t, it->second.second);
        }
    }

    bool is_stable(typename U::type t) const
    {
        typename AbstractCoupledModel::SubModelInternals::const_iterator it =
            submodel_internals.begin();
        bool stable = true;

        while (it != submodel_internals.end() and stable) {
            stable = it->second.first->is_stable(t);
            ++it;
        }
        return stable;
    }

    virtual bool is_updated() const
    { return Externals < T, U >::updated; }

    virtual const Node < U >* get_submodel(unsigned int index) const
    {
        typename AbstractCoupledModel::Submodels::const_iterator it =
            submodels.find(index);

        if (it != submodels.end()) {
            return it->second;
        } else {
            return 0;
        }
    }

    virtual const Node < U >* get_submodel(unsigned int index,
                                           unsigned int rank) const
    {
        typename AbstractCoupledModel::Setsubmodels::const_iterator it =
            setsubmodels.find(index);

        if (it != setsubmodels.end() and it->second.size() > rank) {
            return it->second.at(rank);
        } else {
            return 0;
        }
    }

    virtual std::string path_index(const AbstractModel < U, V >* child,
                                   int& index) const
    {
        typename Setsubmodels::const_iterator it = setsubmodels.begin();
        bool found = false;

        index = -1;
        while (not found and it != setsubmodels.end()) {
            typename std::vector < type* >::const_iterator itm =
                it->second.begin();

            index = 0;
            while (not found and itm != it->second.end()) {
                found = *itm == child;
                if (not found) {
                    ++index;
                    ++itm;
                }
            }
            index = found ? index : -1;
            ++it;
        }
        if (type::parent) {
            int i;
            std::string p = type::parent->path_index(this, i);

            if (i >= 0) {
                return p +
                    "/[" + std::to_string(i) + "]" +
                    boost::core::demangle(typeid(*this).name());
            } else {
                return p + "/" + boost::core::demangle(typeid(*this).name());
            }
        } else {
            return boost::core::demangle(typeid(*this).name());
        }
    }

    virtual void restore(const State < U >& state)
    {
        typename AbstractCoupledModel::Submodels::const_iterator it =
            submodels.begin();

        while (it != submodels.end()) {
            it->second->restore(state.get_substate(it->first));
            ++it;
        }
        States < T, U >::restore(state);
        Internals < T, U >::restore(state);
    }

    virtual void save(State < U >& state) const
    {
        typename AbstractCoupledModel::Submodels::const_iterator it =
            submodels.begin();

        while (it != submodels.end()) {
            State < U > substate;

            it->second->save(substate);
            state.add_substate(it->first, substate);
            ++it;
        }
        States < T, U >::save(state);
        Internals < T, U >::save(state);
    }

    virtual void stable()
    { Externals < T, U >::updated = false; }

protected:
    void change_internal(unsigned int index, double T::* var)
    {
        submodel_internals.erase(index);
        Internals < T, U >::internal(index, var);
    }

    void change_internal(unsigned int index, int T::* var)
    {
        submodel_internals.erase(index);
        Internals < T, U >::internalI(index, var);
    }

    void change_internal(unsigned int index, bool T::* var)
    {
        submodel_internals.erase(index);
        Internals < T, U >::internalB(index, var);
    }

    void internal(unsigned int index, double T::* var)
    { Internals < T, U >::internal(index, var); }

    void internal(unsigned int index, type* model, int sub_index)
    {
        submodel_internals[index] =
            std::pair < type*, int >(model, sub_index);
    }

    template < typename Z >
    void I(std::initializer_list < std::pair < unsigned int,
           Z T::* > > list)
    { Internals < T, U >::I(list); }

    void I(std::initializer_list < Var > internals)
    {
        for (typename std::initializer_list < Var >::iterator it =
                 internals.begin(); it != internals.end(); ++it) {
            submodel_internals[it->index] =
                std::pair < type*, int >(it->model, it->sub_index);
        }
    }

    void submodel(unsigned int index, type* model)
    {
        if (model) {
            submodels[index] = model;
            model->set_parent(this);
        } else {
            submodels[index] = 0;
        }
    }

    void S(std::initializer_list < std::pair <  unsigned int,
           type* > > models)
    {
        for (typename std::initializer_list < std::pair <  unsigned int,
                 type* > >::iterator it =
                 models.begin(); it != models.end(); ++it) {
            submodels[it->first] = it->second;
            it->second->set_parent(this);
        }
    }

    void setsubmodel(unsigned int index, type* model)
    {
        if (setsubmodels.find(index) == setsubmodels.end()) {
            setsubmodels[index] = std::vector < type* >();
        }
        setsubmodels[index].push_back(model);
        model->set_parent(this);
    }

private:
    SubModelInternals submodel_internals;
    Submodels submodels;
    Setsubmodels setsubmodels;
};

} }

#endif
