/**
 * @file test/test.cpp
 * @author See the AUTHORS file
 */

/*
 * Copyright (C) 2012-2016 ULCO http://www.univ-littoral.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define CATCH_CONFIG_MAIN
#include <test/catch.hpp>

#include <artis/kernel/AbstractAtomicModel.hpp>
#include <artis/kernel/AbstractCoupledModel.hpp>
#include <artis/kernel/Builder.hpp>
#include <artis/kernel/Simulator.hpp>
#include <artis/observer/Output.hpp>

#include <artis/utils/DateTime.hpp>
#include <artis/utils/Trace.hpp>

struct GlobalParameters
{ };

struct ModelParameters
{ };

using namespace artis::kernel;
using namespace artis::utils;

using Model = AbstractModel < DoubleTime, ModelParameters >;
template < typename T >
using AtomicModel = AbstractAtomicModel < T, DoubleTime, ModelParameters >;

class AModel : public AtomicModel < AModel >
{
public:
    enum externals { };

    enum internals { DX, BX, IX };

    AModel()
    {
        internal(IX, &AModel::_ix);
        internal(BX, &AModel::_bx);
        internal(DX, &AModel::_dx);
    }

    virtual ~AModel()
    { }

    void compute(double /* t */, bool /* update */)
    {
        ++_ix;
        _bx = not _bx;
        ++_dx;
    }

    void init(double /* t */, const ModelParameters& /* parameters */)
    {
        _ix = 0;
        _bx = false;
        _dx = 0.;
    }

private:
    int _ix;
    bool _bx;
    double _dx;
};

class BModel : public AtomicModel < BModel >
{
public:
    enum externals { IX, BX, DX };
    enum internals { IY, BY, DY };

    BModel()
    {
        // external(IX, &BModel::_ix);
        E < int >({ { IX, &BModel::_ix } });

        external(BX, &BModel::_bx);
        external(DX, &BModel::_dx);

        // internal(IY, &BModel::_iy);
        I < int >({ { IY, &BModel::_iy } });

        internal(BY, &BModel::_by);
        internal(DY, &BModel::_dy);
    }

    virtual ~BModel()
    { }

    void compute(double /* t */, bool update)
    {
        if (update)
            std::cout << "UPDATE" << std::endl;

        _iy = _ix + 1;
        _by = not _bx;
        _dy = _dx + 1;
    }

    void init(double /* t */, const ModelParameters& /* parameters */)
    {
        _iy = 0;
        _by = false;
        _dy = 0.;
    }

private:
    // externals
    int _ix;
    bool _bx;
    double _dx;

    // internals
    int _iy;
    bool _by;
    double _dy;
};

class RootModel :
    public artis::kernel::AbstractCoupledModel < RootModel,
                                                 artis::utils::DoubleTime,
                                                 ModelParameters,
                                                 GlobalParameters >
{
public:
    enum submodels { A, B };

    RootModel()
    {
        S({ { A, &_a }, { B, &_b } });
    }

    virtual ~RootModel()
    { }

    void compute(double t, bool /* update */)
    {
        _a(t);

        _b.put < double >(t, BModel::DX, _a.get < double >(t, AModel::DX));
        _b.put < int >(t, BModel::IX, _a.get < int >(t, AModel::IX));
        _b.put < bool >(t, BModel::BX, _a.get < bool >(t, AModel::BX));
        _b(t);
    }

    void init(double t, const ModelParameters& parameters)
    {
        _a.init(t, parameters);
        _b.init(t, parameters);
    }

private:
    AModel _a;
    BModel _b;
};

typedef artis::kernel::Simulator < RootModel,
                                   artis::utils::DoubleTime,
                                   ModelParameters,
                                   GlobalParameters > ASimulator;

typedef artis::kernel::Builder < RootModel,
                                 artis::utils::DoubleTime,
                                 ModelParameters,
                                 GlobalParameters > ABuilder;

class AView : public artis::observer::View < artis::utils::DoubleTime,
                                             ModelParameters >
{
public:
    AView()
    {
        selector("A::I", INT, { RootModel::A, AModel::IX });
        selector("A::B", BOOL, { RootModel::A, AModel::BX });
        selector("A::D", DOUBLE, { RootModel::A, AModel::DX });
        selector("B::I", INT, { RootModel::B, BModel::IY });
        selector("B::B", BOOL, { RootModel::B, BModel::BY });
        selector("B::D", DOUBLE, { RootModel::B, BModel::DY });
    }

    virtual ~AView()
    { }
};

typedef artis::observer::Output < artis::utils::DoubleTime,
                                  ModelParameters > AnOutput;

TEST_CASE("Simulator_tests", "simple")
{
    GlobalParameters globalParameters;
    ModelParameters modelParameters;
    ABuilder builder("{ \"type\": \"RootModel\", "                      \
                     "\"submodels\": [ "                                \
                     "{ \"type\": \"AModel\", \"internals\": [] }, "    \
                     "{ \"type\": \"BModel\", \"internals\": [] } "     \
                     "] }");
    ASimulator simulator(new RootModel, globalParameters);

    simulator.attachView("Root", new AView);

    artis::utils::Trace < artis::utils::DoubleTime >::trace().clear();

    simulator.init(0, modelParameters);
    simulator.run(artis::utils::DateTime::toJulianDayNumber(2016, 1, 1),
                  artis::utils::DateTime::toJulianDayNumber(2016, 1, 10));

    AnOutput output(simulator.observer());

    output();
}
